# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import sqlalchemy as sa

from tacker.db import model_base


class LccnSubscriptionV2(model_base.BASE):
    """Type: LccnSubscription

    NFV-SOL 003
    - v3.3.1 5.5.2.16 (API version: 2.0.0)
    """

    __tablename__ = 'LccnSubscriptionV2'
    id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    filter = sa.Column(sa.JSON(), nullable=True)
    callbackUri = sa.Column(sa.String(255), nullable=False)
    # NOTE: 'authentication' attribute is not included in the
    #       original 'LccnSubscription' data type definition.
    authentication = sa.Column(sa.JSON(), nullable=True)
    verbosity = sa.Column(sa.Enum('FULL', 'SHORT', create_constraint=True,
                                  validate_strings=True), nullable=False)


class VnfInstanceV2(model_base.BASE):
    """Type: VnfInstance

    NFV-SOL 003
    - v3.3.1 5.5.2.2 (API version: 2.0.0)
    """

    __tablename__ = 'VnfInstanceV2'
    id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    vnfInstanceName = sa.Column(sa.String(255), nullable=True)
    vnfInstanceDescription = sa.Column(sa.Text(), nullable=True)
    vnfdId = sa.Column(sa.String(255), nullable=False)
    vnfProvider = sa.Column(sa.String(255), nullable=False)
    vnfProductName = sa.Column(sa.String(255), nullable=False)
    vnfSoftwareVersion = sa.Column(sa.String(255), nullable=False)
    vnfdVersion = sa.Column(sa.String(255), nullable=False)
    vnfConfigurableProperties = sa.Column(sa.JSON(), nullable=True)
    vimConnectionInfo = sa.Column(sa.JSON(), nullable=True)
    instantiationState = sa.Column(sa.Enum(
        'NOT_INSTANTIATED', 'INSTANTIATED', create_constraint=True,
        validate_strings=True), nullable=False)
    instantiatedVnfInfo = sa.Column(sa.JSON(), nullable=True)
    metadata__ = sa.Column("metadata", sa.JSON(), nullable=True)
    extensions = sa.Column(sa.JSON(), nullable=True)


class VnfLcmOpOccV2(model_base.BASE):
    """Type: VnfLcmOpOcc

    NFV-SOL 003
    - v3.3.1 5.5.2.13 (API version: 2.0.0)
    """

    __tablename__ = 'VnfLcmOpOccV2'
    id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    operationState = sa.Column(sa.Enum(
        'STARTING', 'PROCESSING', 'COMPLETED', 'FAILED_TEMP',
        'FAILED', 'ROLLING_BACK', 'ROLLED_BACK',
        create_constraint=True, validate_strings=True), nullable=False)
    stateEnteredTime = sa.Column(sa.DateTime(), nullable=False)
    startTime = sa.Column(sa.DateTime(), nullable=False)
    vnfInstanceId = sa.Column(sa.String(255), nullable=False)
    grantId = sa.Column(sa.String(255), nullable=True)
    operation = sa.Column(sa.Enum(
        'INSTANTIATE', 'SCALE', 'SCALE_TO_LEVEL', 'CHANGE_FLAVOUR',
        'TERMINATE', 'HEAL', 'OPERATE', 'CHANGE_EXT_CONN',
        'MODIFY_INFO', 'CREATE_SNAPSHOT', 'REVERT_TO_SNAPSHOT',
        'CHANGE_VNFPKG', create_constraint=True, validate_strings=True),
        nullable=False)
    isAutomaticInvocation = sa.Column(sa.Boolean, nullable=False)
    operationParams = sa.Column(sa.JSON(), nullable=True)
    isCancelPending = sa.Column(sa.Boolean, nullable=False)
    cancelMode = sa.Column(sa.Enum(
        'GRACEFUL', 'FORCEFUL', create_constraint=True, validate_strings=True),
        nullable=True)
    error = sa.Column(sa.JSON(), nullable=True)
    resourceChanges = sa.Column(sa.JSON(), nullable=True)
    changedInfo = sa.Column(sa.JSON(), nullable=True)
    changedExtConnectivity = sa.Column(sa.JSON(), nullable=True)
    modificationsTriggeredByVnfPkgChange = sa.Column(sa.JSON(), nullable=True)
    vnfSnapshotInfoId = sa.Column(sa.String(255), nullable=True)


class GrantV1(model_base.BASE):
    """Type: Grant

    NFV-SOL 003
    - v3.3.1 9.5.2.3 (API version: 1.4.0)
    """

    __tablename__ = 'GrantV1'
    id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    vnfInstanceId = sa.Column(sa.String(255), nullable=False)
    vnfLcmOpOccId = sa.Column(sa.String(255), nullable=False)
    vimConnectionInfo = sa.Column(sa.JSON(), nullable=True)
    zones = sa.Column(sa.JSON(), nullable=True)
    zoneGroups = sa.Column(sa.JSON(), nullable=True)
    addResources = sa.Column(sa.JSON(), nullable=True)
    tempResources = sa.Column(sa.JSON(), nullable=True)
    removeResources = sa.Column(sa.JSON(), nullable=True)
    updateResources = sa.Column(sa.JSON(), nullable=True)
    vimAssets = sa.Column(sa.JSON(), nullable=True)
    extVirtualLinks = sa.Column(sa.JSON(), nullable=True)
    extManagedVirtualLinks = sa.Column(sa.JSON(), nullable=True)
    additionalParams = sa.Column(sa.JSON(), nullable=True)


class GrantRequestV1(model_base.BASE):
    """Type: GrantRequest

    NFV-SOL 003
    - v3.3.1 9.5.2.2 (API version: 1.4.0)
    """

    __tablename__ = 'GrantRequestV1'
    vnfInstanceId = sa.Column(sa.String(255), nullable=False)
    vnfLcmOpOccId = sa.Column(sa.String(255), nullable=False, primary_key=True)
    vnfdId = sa.Column(sa.String(255), nullable=False)
    dstVnfdId = sa.Column(sa.String(255), nullable=True)
    flavourId = sa.Column(sa.String(255), nullable=True)
    operation = sa.Column(sa.Enum(
        'INSTANTIATE', 'SCALE', 'SCALE_TO_LEVEL', 'CHANGE_FLAVOUR',
        'TERMINATE', 'HEAL', 'OPERATE', 'CHANGE_EXT_CONN',
        'CREATE_SNAPSHOT', 'REVERT_TO_SNAPSHOT', 'CHANGE_VNFPKG',
        create_constraint=True, validate_strings=True),
        nullable=False)
    isAutomaticInvocation = sa.Column(sa.Boolean, nullable=False)
    instantiationLevelId = sa.Column(sa.String(255), nullable=True)
    addResources = sa.Column(sa.JSON(), nullable=True)
    tempResources = sa.Column(sa.JSON(), nullable=True)
    removeResources = sa.Column(sa.JSON(), nullable=True)
    updateResources = sa.Column(sa.JSON(), nullable=True)
    placementConstraints = sa.Column(sa.JSON(), nullable=True)
    vimConstraints = sa.Column(sa.JSON(), nullable=True)
    additionalParams = sa.Column(sa.JSON(), nullable=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fixtures for Tacker unit tests."""
# NOTE(bhagyashris): This is needed for importing from fixtures.

import warnings

import fixtures as pyfixtures


class WarningsFixture(pyfixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super(WarningsFixture, self).setUp()
        # NOTE(bhagyashris): user/tenant is deprecated in oslo.context
        # so don't let anything new use it
        warnings.filterwarnings(
            'error', message="Property '.*' has moved to '.*'")

        self.addCleanup(warnings.resetwarnings)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import random
import time

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from tacker.objects import fields
from tacker.tests.functional import base
from tacker.tests import utils
from tacker.vnfm.infra_drivers.openstack import constants as infra_cnst


VNF_PACKAGE_UPLOAD_TIMEOUT = 300
VNF_INSTANTIATE_TIMEOUT = 600
VNF_TERMINATE_TIMEOUT = 600
VNF_HEAL_TIMEOUT = 600
VNF_CHANGE_EXT_CONN_TIMEOUT = 600
RETRY_WAIT_TIME = 5


def get_ext_managed_virtual_link(id, vl_desc_id, resource_id):
    return [{"id": id, "vnfVirtualLinkDescId": vl_desc_id,
            "resourceId": resource_id}]


def generate_mac_address():
    """Generate an Ethernet MAC address."""
    mac = [0xfa, 0x16, 0x3e,
           random.randint(0x00, 0xff),
           random.randint(0x00, 0xff),
           random.randint(0x00, 0xff)]
    return ':'.join(map(lambda x: "%02x" % x, mac))


def generate_ip_addresses(
        type_='IPV4',
        fixed_addresses=None,
        subnet_id=None):
    if fixed_addresses:
        ip_addr = {
            'type': type_,
            'fixedAddresses': fixed_addresses
        }
        if subnet_id:
            ip_addr.update({'subnetId': subnet_id})
    return [ip_addr]


def get_ext_cp_with_external_link_port(nw_resource_id, port_uuid):
    ext_cp = {
        "id": "external_network",
        "resourceId": nw_resource_id,
        "extCps": [{
            "cpdId": "CP2",
            "cpConfig": [{
                "linkPortId": "413f4e46-21cf-41b1-be0f-de8d23f76cfe",
                "cpProtocolData": [{
                    "layerProtocol": "IP_OVER_ETHERNET"
                }]
            }]
        }],
        "extLinkPorts": [{
            "id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe",
            "resourceHandle": {
                "resourceId": port_uuid,
                "vimLevelResourceType": "LINKPORT"
            }
        }]
    }
    return ext_cp


def get_ext_cp_with_fixed_address(nw_resource_id, fixed_addresses, subnet_id):
    ext_cp = {
        "id": "external_network",
        "resourceId": nw_resource_id,
        "extCps": [{
            "cpdId": "CP2",
            "cpConfig": [{
                "cpProtocolData": [{
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": generate_ip_addresses(
                            fixed_addresses=fixed_addresses,
                            subnet_id=subnet_id)
                    }
                }]
            }]
        }]
    }
    return ext_cp


def get_external_virtual_links(net_0_resource_id, net_mgmt_resource_id,
        port_uuid, fixed_addresses=None, subnet_id=None):
    ext_vl = [
        {
            "id": "net0",
            "resourceId": net_0_resource_id,
            "extCps": [{
                "cpdId": "CP1",
                "cpConfig": [{
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "macAddress": generate_mac_address()
                        }
                    }]
                }]
            }]
        }
    ]
    if fixed_addresses:
        ext_cp = get_ext_cp_with_fixed_address(
            net_mgmt_resource_id, fixed_addresses, subnet_id)
    else:
        ext_cp = get_ext_cp_with_external_link_port(
            net_mgmt_resource_id, port_uuid)
    ext_vl.append(ext_cp)

    return ext_vl


def _create_and_upload_vnf_package(tacker_client, csar_package_name,
        user_defined_data):
    # create vnf package
    body = jsonutils.dumps({"userDefinedData": user_defined_data})
    resp, vnf_package = tacker_client.do_request(
        '/vnfpkgm/v1/vnf_packages', "POST", body=body)

    # upload vnf package
    csar_package_path = "../../../etc/samples/etsi/nfv/%s" % csar_package_name
    file_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
        csar_package_path))

    # Generating unique vnfd id. This is required when multiple workers
    # are running concurrently. The call below creates a new temporary
    # CSAR with unique vnfd id.
    file_path, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)

    with open(file_path, 'rb') as file_object:
        resp, resp_body = tacker_client.do_request(
            '/vnfpkgm/v1/vnf_packages/{id}/package_content'.format(
                id=vnf_package['id']),
            "PUT", body=file_object, content_type='application/zip')

    # wait for onboard
    timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
    start_time = int(time.time())
    show_url = os.path.join('/vnfpkgm/v1/vnf_packages', vnf_package['id'])
    vnfd_id = None
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if body['onboardingState'] == "ONBOARDED":
            vnfd_id = body['vnfdId']
            break

        if ((int(time.time()) - start_time) > timeout):
            raise Exception("Failed to onboard vnf package")

        time.sleep(1)

    # remove temporarily created CSAR file
    os.remove(file_path)
    return vnf_package['id'], vnfd_id


class VnfLcmTest(base.BaseTackerTest):

    @classmethod
    def setUpClass(cls):
        cls.tacker_client = base.BaseTackerTest.tacker_http_client()

        cls.vnf_package_1, cls.vnfd_id_1 = _create_and_upload_vnf_package(
            cls.tacker_client, "vnflcm1", {"key": "sample_1_functional"})

        cls.vnf_package_2, cls.vnfd_id_2 = _create_and_upload_vnf_package(
            cls.tacker_client, "vnflcm2", {"key": "sample_2_functional"})

        cls.vnf_package_3, cls.vnfd_id_3 = _create_and_upload_vnf_package(
            cls.tacker_client, "vnflcm3", {"key": "sample_3_functional"})

        super(VnfLcmTest, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        # Update vnf package operational state to DISABLED
        update_req_body = jsonutils.dumps({
            "operationalState": "DISABLED"})
        base_path = "/vnfpkgm/v1/vnf_packages"
        for package_id in [cls.vnf_package_1, cls.vnf_package_2,
                cls.vnf_package_3]:
            resp, resp_body = cls.tacker_client.do_request(
                '{base_path}/{id}'.format(id=package_id,
                                          base_path=base_path),
                "PATCH", content_type='application/json', body=update_req_body)

            # Delete vnf package
            url = '/vnfpkgm/v1/vnf_packages/%s' % package_id
            cls.tacker_client.do_request(url, "DELETE")

        super(VnfLcmTest, cls).tearDownClass()

    def setUp(self):
        super(VnfLcmTest, self).setUp()
        self.base_url = "/vnflcm/v1/vnf_instances"
        self.base_vnf_lcm_op_occs_url = "/vnflcm/v1/vnf_lcm_op_occs"

        vim_list = self.client.list_vims()
        if not vim_list:
            self.skipTest("Vims are not configured")

        vim_id = 'VIM0'
        vim = self.get_vim(vim_list, vim_id)
        if not vim:
            self.skipTest("Default VIM '%s' is missing" % vim_id)
        self.vim_id = vim['id']

    def _instantiate_vnf_request(self, flavour_id,
            instantiation_level_id=None, vim_id=None, ext_vl=None,
            ext_managed_vl=None):
        request_body = {"flavourId": flavour_id}

        if instantiation_level_id:
            request_body["instantiationLevelId"] = instantiation_level_id

        if ext_managed_vl:
            request_body["extManagedVirtualLinks"] = ext_managed_vl

        if ext_vl:
            request_body["extVirtualLinks"] = ext_vl

        if vim_id:
            request_body["vimConnectionInfo"] = [
                {"id": uuidutils.generate_uuid(),
                 "vimId": vim_id,
                 "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2"}]

        return request_body

    def _create_vnf_instance(self, vnfd_id, vnf_instance_name=None,
            vnf_instance_description=None):
        request_body = {'vnfdId': vnfd_id}
        if vnf_instance_name:
            request_body['vnfInstanceName'] = vnf_instance_name

        if vnf_instance_description:
            request_body['vnfInstanceDescription'] = vnf_instance_description

        resp, response_body = self.http_client.do_request(
            self.base_url, "POST", body=jsonutils.dumps(request_body))
        return resp, response_body

    def _delete_wait_vnf_instance(self, id):
        timeout = VNF_TERMINATE_TIMEOUT
        url = os.path.join(self.base_url, id)
        start_time = int(time.time())
        while True:
            resp, body = self.http_client.do_request(url, "DELETE")
            if 204 == resp.status_code:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = "Failed to delete vnf instance %s"
                self.fail(error % id)

            time.sleep(RETRY_WAIT_TIME)

    def _delete_vnf_instance(self, id):
        self._delete_wait_vnf_instance(id)

        # verify vnf instance is deleted
        url = os.path.join(self.base_url, id)
        resp, body = self.http_client.do_request(url, "GET")
        self.assertEqual(404, resp.status_code)

    def _show_vnf_instance(self, id, expected_result=None):
        show_url = os.path.join(self.base_url, id)
        resp, vnf_instance = self.http_client.do_request(show_url, "GET")
        self.assertEqual(200, resp.status_code)

        if expected_result:
            self.assertDictSupersetOf(expected_result, vnf_instance)

        return vnf_instance

    def _list_vnf_instances(self):
        resp, vnf_instances = self.http_client.do_request(self.base_url, "GET")
        self.assertEqual(200, resp.status_code)
        return vnf_instances

    def _stack_update_wait(self, stack_id, expected_status,
            timeout=VNF_HEAL_TIMEOUT):
        start_time = int(time.time())
        while True:
            stack = self.h_client.stacks.get(stack_id)
            if stack.stack_status == expected_status:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = ("Stack %(id)s status is %(current)s, expected status "
                        "should be %(expected)s")
                self.fail(error % {"id": stack_id, "current": stack.status,
                    "expected": expected_status})

            time.sleep(RETRY_WAIT_TIME)

    def _vnf_instance_wait(self, id,
            instantiation_state=fields.VnfInstanceState.INSTANTIATED,
            timeout=VNF_INSTANTIATE_TIMEOUT):
        show_url = os.path.join(self.base_url, id)
        start_time = int(time.time())
        while True:
            resp, body = self.http_client.do_request(show_url, "GET")
            if body['instantiationState'] == instantiation_state:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = ("Vnf instance %(id)s status is %(current)s, "
                         "expected status should be %(expected)s")
                self.fail(error % {"id": id,
                    "current": body['instantiationState'],
                    "expected": instantiation_state})

            time.sleep(RETRY_WAIT_TIME)

    def _create_network(self, neutron_client, network_name):
        net = neutron_client.create_network(
            {'network': {'name': "network-%s" % uuidutils.generate_uuid()}})
        net_id = net['network']['id']
        self.addCleanup(neutron_client.delete_network, net_id)
        return net_id

    def _create_subnet(self, neutron_client, network_id):
        body = {'subnet': {'network_id': network_id,
                'name': "subnet-%s" % uuidutils.generate_uuid(),
                'cidr': "22.22.0.0/24",
                'ip_version': 4,
                'gateway_ip': '22.22.0.1',
                "enable_dhcp": True}}
        subnet = neutron_client.create_subnet(body=body)["subnet"]
        self.addCleanup(neutron_client.delete_subnet, subnet['id'])
        return subnet['id']

    def _create_port(self, neutron_client, network_id):
        body = {'port': {'network_id': network_id}}
        port = neutron_client.create_port(body=body)["port"]
        self.addCleanup(neutron_client.delete_port, port['id'])
        return port['id']

    def _instantiate_vnf_instance(self, id, request_body):
        url = os.path.join(self.base_url, id, "instantiate")
        resp, body = self.http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        self._vnf_instance_wait(id)

    def _terminate_vnf_instance(self, id, request_body):
        url = os.path.join(self.base_url, id, "terminate")
        resp, body = self.http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        timeout = request_body.get('gracefulTerminationTimeout')
        start_time = int(time.time())

        self._vnf_instance_wait(id,
            instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=VNF_TERMINATE_TIMEOUT)

        # If gracefulTerminationTimeout is set, check whether vnf
        # instantiation_state is set to NOT_INSTANTIATED after
        # gracefulTerminationTimeout seconds.
        if timeout and int(time.time()) - start_time < timeout:
            self.fail("Vnf is terminated before graceful termination "
                      "timeout period")

    def _heal_vnf_instance(self, vnf_instance, request_body,
            expected_stack_status=infra_cnst.STACK_UPDATE_COMPLETE):
        url = os.path.join(self.base_url, vnf_instance['id'], "heal")
        resp, body = self.http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        stack = self.h_client.stacks.get(vnf_instance['vnfInstanceName'])
        # Wait until tacker heals the stack resources as requested in
        # in the heal request
        self._stack_update_wait(stack.id, expected_stack_status)

    def _heal_sol_003_vnf_instance(self, vnf_instance, request_body):
        url = os.path.join(self.base_url, vnf_instance['id'], "heal")
        resp, body = self.http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        # If healing is done without vnfc components, it will delete the
        # stack and create a new one. So wait until vnf is deleted and then
        # wait until a new stack is created using vnfInstanceName and once
        # the stack is created, wait until it's status becomes
        # CREATE_COMPLETE.
        stack = self.h_client.stacks.get(vnf_instance['vnfInstanceName'])
        self._stack_update_wait(stack.id, infra_cnst.STACK_DELETE_COMPLETE)
        start_time = int(time.time())
        timeout = VNF_INSTANTIATE_TIMEOUT
        while True:
            try:
                stack = self.h_client.stacks.get(
                    vnf_instance['vnfInstanceName'])
                if stack.stack_status == infra_cnst.STACK_CREATE_COMPLETE:
                    break
            except Exception:
                pass

            if ((int(time.time()) - start_time) > timeout):
                self.fail("Failed to heal vnf during instantiation")

    def _get_server(self, server_id):
        try:
            self.novaclient().servers.get(server_id)
        except Exception:
            self.fail("Failed to get vdu resource %s id" % server_id)

    def _verify_vnfc_resource_info(self, vnf_instance_old,
            vnf_instance_current, vdu_count):
        vnfc_resource_info_old = (vnf_instance_old['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        vnfc_resource_info_current = (vnf_instance_current
            ['instantiatedVnfInfo']['vnfcResourceInfo'])
        for index in range(vdu_count):
            # compare computeResource resourceId is different
            vdu_resource_id_old = (vnfc_resource_info_old[index]
                ['computeResource']['resourceId'])
            vdu_resource_id_current = (vnfc_resource_info_current[index]
                ['computeResource']['resourceId'])
            self.assertNotEqual(vdu_resource_id_old, vdu_resource_id_current)

            # Now check whether vdus are healed properly and servers exists
            # in nova.
            self._get_server(vdu_resource_id_current)

    def _change_ext_conn_vnf_request(self, vim_id=None, ext_vl=None):
        request_body = {}
        if ext_vl:
            request_body["extVirtualLinks"] = ext_vl

        if vim_id:
            request_body["vimConnectionInfo"] = [
                {"id": uuidutils.generate_uuid(),
                 "vimId": vim_id,
                 "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2"}]

        return request_body

    def _change_ext_conn_vnf_instance(self, vnf_instance, request_body,
            expected_stack_status=infra_cnst.STACK_UPDATE_COMPLETE):
        url = os.path.join(self.base_url, vnf_instance['id'],
            "change_ext_conn")
        resp, body = self.http_client.do_request(url, "POST",
            body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        stack = self.h_client.stacks.get(vnf_instance['vnfInstanceName'])
        # Wait until tacker changes the stack resources as requested
        # in the change_ext_conn request
        self._stack_update_wait(stack.id, expected_stack_status,
                                VNF_CHANGE_EXT_CONN_TIMEOUT)

    def _get_heat_stack(self, vnf_instance_id, stack_name):
        heatclient = self.heatclient()
        try:
            stacks = heatclient.stacks.list()
        except Exception:
            return None

        target_stakcs = list(
            filter(
                lambda x: x.stack_name == stack_name,
                stacks))

        if len(target_stakcs) == 0:
            return None

        return target_stakcs[0]

    def _get_heat_resource_info(self, stack_id, nested_depth=0,
            resource_name=None):
        heatclient = self.heatclient()
        try:
            if resource_name is None:
                resources = heatclient.resources.list(stack_id,
                                         nested_depth=nested_depth)
            else:
                resources = heatclient.resources.get(stack_id,
                                         resource_name)
        except Exception:
            return None
        return resources

    def _get_fixed_ips(self, vnf_instance, request_body):
        vnf_instance_id = vnf_instance['id']
        vnf_instance_name = vnf_instance['vnfInstanceName']
        res_name = None
        for extvirlink in request_body['extVirtualLinks']:
            if 'extCps' not in extvirlink:
                continue
            for extcps in extvirlink['extCps']:
                if 'cpdId' in extcps:
                    if res_name is None:
                        res_name = list()
                    res_name.append(extcps['cpdId'])
                    break
        if res_name is None:
            return []

        stack = self._get_heat_stack(vnf_instance_id,
                                     vnf_instance_name)
        stack_id = stack.id

        stack_resource = self._get_heat_resource_info(
            stack_id, nested_depth=2)

        releations = dict()
        for elmt in stack_resource:
            if elmt.resource_type != 'OS::Neutron::Port':
                continue
            if elmt.resource_name not in res_name:
                continue
            parent = getattr(elmt, 'parent_resource', None)
            releations[parent] = elmt.resource_name

        details = dict()
        for (parent_name, resource_name) in releations.items():
            for elmt in stack_resource:
                if parent_name is None:
                    detail_stack = self._get_heat_resource_info(
                        stack_id, resource_name=resource_name)
                elif parent_name != elmt.resource_name:
                    continue
                else:
                    detail_stack = self._get_heat_resource_info(
                        elmt.physical_resource_id, resource_name=resource_name)
                details[resource_name] = detail_stack

        ans_list = list()
        for detail in details.values():
            ans_list.append(detail.attributes['fixed_ips'])

        return ans_list

    def test_create_show_delete_vnf_instance(self):
        """Create, show and delete a vnf instance."""

        # Create vnf instance
        vnf_instance_name = "Test-VNf-Instance"
        vnf_instance_description = "Sample VNF for LCM Testing"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_1,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        expected_result = {
            "instantiationState": fields.VnfInstanceState.NOT_INSTANTIATED,
            "vnfInstanceName": vnf_instance_name,
            "vnfInstanceDescription": vnf_instance_description
        }
        self._show_vnf_instance(vnf_instance['id'], expected_result)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_list_vnf_instances(self):
        """Create vnf instances and check list API display those vnfs."""

        # Create vnf instance 01 and don't instantiate this one.
        vnf_instance_name = "List-VNF-Instance-0"
        resp, vnf_instance_0 = self._create_vnf_instance(self.vnfd_id_1,
                vnf_instance_name=vnf_instance_name)

        self.assertIsNotNone(vnf_instance_0['id'])
        self.assertEqual(201, resp.status_code)

        self.addCleanup(self._delete_vnf_instance, vnf_instance_0['id'])

        # Create vnf instance 02 and instantiate this one.
        vnf_instance_name = "List-VNF-Instance-1"
        resp, vnf_instance_1 = self._create_vnf_instance(self.vnfd_id_1,
                vnf_instance_name=vnf_instance_name)

        self.assertIsNotNone(vnf_instance_1['id'])
        self.assertEqual(201, resp.status_code)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id)

        self._instantiate_vnf_instance(vnf_instance_1['id'], request_body)
        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self.addCleanup(self._delete_vnf_instance, vnf_instance_1['id'])
        self.addCleanup(self._terminate_vnf_instance, vnf_instance_1['id'],
                        terminate_req_body)

        # List vnf instances to check if first one is in NOT_INSTANTIATED
        # state and the second one is INSTANTIATED
        vnf_instances = self._list_vnf_instances()
        for vnf_instance in vnf_instances:
            if vnf_instance['id'] == vnf_instance_0['id']:
                self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                        vnf_instance['instantiationState'])
            elif vnf_instance['id'] == vnf_instance_1['id']:
                self.assertEqual(fields.VnfInstanceState.INSTANTIATED,
                        vnf_instance['instantiationState'])

    def test_instantiate_vnf_with_flavour(self):
        """Test instantiation and heal API without instantiation level

        This test will instantiate vnf using flavour. Heal API will be invoked
        by passing vnfcInstanceId parameter in the request body as per SOL002
        HealVnfRequest.
        """

        # Create vnf instance
        vnf_instance_name = "vnf_with_flavour-%s" % uuidutils.generate_uuid()
        vnf_instance_description = "vnf with instantiation level and no ext vl"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_1,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(1, vdu_count)

        # heal as per SOL002 API check, i.e. pass vnfcInstanceId in the
        # HealVnfRequest.
        vnfc_resource_info = (vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        vnfInstanceIds = [vnfc_res_info['id'] for vnfc_res_info in
            vnfc_resource_info]

        heal_request_body = {
            "cause": "Heal as per SOL002 API check",
            "vnfcInstanceId": vnfInstanceIds
        }

        self._heal_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 1)

        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_instantiate_vnf_with_instantiation_level(self):
        """Test instantiation and heal API with instantiation level

        This test will instantiate vnf with instantiation level. Heal API
        will be invoked by passing vnfcInstanceId parameter in the request
        body as per SOL002 HealVnfRequest.
        """

        # Create vnf instance
        vnf_instance_name = "vnf_with_instantiation_level-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf with instantiation level 2"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_2,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        request_body = self._instantiate_vnf_request("simple",
                instantiation_level_id="instantiation_level_2",
                vim_id=self.vim_id)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(3, vdu_count)

        # heal as per SOL002 API check, i.e.vnfcInstanceId is passed in
        # the HealVnfRequest.
        vnfc_resource_info = (vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        vnfInstanceIds = [vnfc_res_info['id'] for vnfc_res_info in
            vnfc_resource_info]

        heal_request_body = {
            "cause": "Heal as per SOL002 API check",
            "vnfcInstanceId": vnfInstanceIds
        }

        self._heal_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 3)

        # Terminate vnf forcefully
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.FORCEFUL
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_instantiate_vnf_with_ext_vl_and_ext_managed_vl(self):
        """Test instantiation vnf with external virtual links

        This test will instantiate vnf with external virtual links and
        external managed virtual links. Heal API will be invoked by
        passing vnfcInstanceId parameter in the request body as per SOL002
        HealVnfRequest.
        """

        # Create vnf instance
        vnf_instance_name = "vnf_with_ext_vl_and_ext_managed_vl-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf_with_ext_vl_and_ext_managed_vl"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_3,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        neutron_client = self.neutronclient()
        net = neutron_client.list_networks()
        networks = {}
        for network in net['networks']:
            networks[network['name']] = network['id']

        net1_id = networks.get('net1')
        if not net1_id:
            self.fail("net1 network is not available")

        net0_id = networks.get('net0')
        if not net0_id:
            self.fail("net0 network is not available")

        net_mgmt_id = networks.get('net_mgmt')
        if not net_mgmt_id:
            self.fail("net_mgmt network is not available")

        ext_managed_vl = get_ext_managed_virtual_link("net1", "VL3",
            net1_id)

        network_uuid = self._create_network(neutron_client,
                                            "external_network")
        self._create_subnet(neutron_client, network_uuid)
        port_uuid = self._create_port(neutron_client, network_uuid)
        ext_vl = get_external_virtual_links(net0_id, net_mgmt_id,
                                            port_uuid)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id, ext_vl=ext_vl, ext_managed_vl=ext_managed_vl)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(1, vdu_count)

        # heal as per SOL002 API check, i.e.vnfcInstanceId is passed in
        # the HealVnfRequest.
        vnfc_resource_info = (vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        vnfInstanceIds = [vnfc_res_info['id'] for vnfc_res_info in
            vnfc_resource_info]

        heal_request_body = {
            "cause": "Heal as per SOL002 API check",
            "vnfcInstanceId": vnfInstanceIds
        }

        self._heal_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 1)

        # Terminate vnf forcefully
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.FORCEFUL
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_heal_vnf_sol_003_with_flavour(self):
        """Test heal API as per SOL 003 for VNF created with flavor

        This test will instantiate vnf using flavour. Heal API will be invoked
        as per SOL003 i.e. without passing vnfcInstanceId, so that the entire
        vnf is healed which includes VDU/CP/VL/STORAGE.
        """

        # Create vnf instance
        vnf_instance_name = "heal_vnf_sol_003_with_flavour-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf with instantiation level and no ext vl"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_1,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(1, vdu_count)

        # heal as per SOL003 API check, i.e. without passing vnfcInstanceId in
        # the HealVnfRequest.
        heal_request_body = {
            "cause": "Heal as per SOL003 API check",
        }

        self._heal_sol_003_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 1)

        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_heal_vnf_sol_003_with_instantiation_level(self):
        """Test heal as per SOL003 for VNF created with instantiation level

        This test will instantiate vnf with instantiation level. Heal API will
        be invoked as per SOL003 i.e. without passing vnfcInstanceId, so that
        the entire vnf is healed which includes VDU/CP/VL/STORAGE.
        """

        # Create vnf instance
        vnf_instance_name = "heal_vnf_sol_003_with_instantiation_level-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf with instantiation level 2"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_2,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        request_body = self._instantiate_vnf_request("simple",
                instantiation_level_id="instantiation_level_2",
                vim_id=self.vim_id)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(3, vdu_count)

        # heal as per SOL003 API check, i.e. without passing vnfcInstanceId
        # in the HealVnfRequest.
        heal_request_body = {
            "cause": "Heal as per SOL003 API check",
        }

        self._heal_sol_003_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 3)

        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_heal_vnf_sol_003_ext_vl_and_ext_managed_vl(self):
        """Test heal vnf as per SOL003 with vnf created using external vl.

        This test will instantiate vnf with external virtual links and
        external managed virtual links. Heal API will be invoked as per SOL003
        i.e. without passing vnfcInstanceId, so that the entire vnf is healed
        which includes VDU/CP/VL/STORAGE.
        """

        # Create vnf instance
        vnf_instance_name = "vnf_with_ext_vl_and_ext_managed_vl-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf_with_ext_vl_and_ext_managed_vl"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_3,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        neutron_client = self.neutronclient()
        net = neutron_client.list_networks()
        networks = {}
        for network in net['networks']:
            networks[network['name']] = network['id']

        net1_id = networks.get('net1')
        if not net1_id:
            self.fail("net1 network is not available")

        net0_id = networks.get('net0')
        if not net0_id:
            self.fail("net0 network is not available")

        net_mgmt_id = networks.get('net_mgmt')
        if not net_mgmt_id:
            self.fail("net_mgmt network is not available")

        ext_managed_vl = get_ext_managed_virtual_link("net1", "VL3",
            net1_id)

        network_uuid = self._create_network(neutron_client,
                                            "external_network")
        self._create_subnet(neutron_client, network_uuid)
        port_uuid = self._create_port(neutron_client, network_uuid)
        ext_vl = get_external_virtual_links(net0_id, net_mgmt_id,
                                            port_uuid)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id, ext_vl=ext_vl, ext_managed_vl=ext_managed_vl)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(1, vdu_count)

        # heal as per SOL003 API check, i.e. without passing vnfcInstanceId
        # in the HealVnfRequest.
        heal_request_body = {
            "cause": "Heal as per SOL003 API check",
        }

        self._heal_sol_003_vnf_instance(vnf_instance, heal_request_body)

        # NOTE(tpatil) Wait for sometime as it takes a while to update
        # vnfcResourceInfo after the stack status becomes UPDATE_COMPLETE.
        # There is no intermediate status set to VNF which can be used here
        # to confirm healing action is completed successfully.
        time.sleep(20)

        vnf_instance_current = self._show_vnf_instance(vnf_instance['id'])
        self._verify_vnfc_resource_info(vnf_instance, vnf_instance_current, 1)

        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])

    def test_inst_chgextconn_term(self):
        """Test change external vnf connectivity.

        This test will instantiate vnf with external virtual link and
        change the IP address on virtual link.
        """

        # Create vnf instance
        vnf_instance_name = "vnf_with_ext_vl_and_ext_managed_vl-%s" % \
            uuidutils.generate_uuid()
        vnf_instance_description = "vnf_with_ext_vl_and_ext_managed_vl"
        resp, vnf_instance = self._create_vnf_instance(self.vnfd_id_3,
                vnf_instance_name=vnf_instance_name,
                vnf_instance_description=vnf_instance_description)

        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        neutron_client = self.neutronclient()
        net = neutron_client.list_networks()
        networks = {}
        for network in net['networks']:
            networks[network['name']] = network['id']
        subnet_list = neutron_client.list_subnets()
        subnets = {}
        for subnet in subnet_list['subnets']:
            subnets[subnet['name']] = subnet['id']

        net1_id = networks.get('net1')
        if not net1_id:
            self.fail("net1 network is not available")

        net0_id = networks.get('net0')
        if not net0_id:
            self.fail("net0 network is not available")

        net_mgmt_id = networks.get('net_mgmt')
        if not net_mgmt_id:
            self.fail("net_mgmt network is not available")

        subnet_mgmt_id = subnets.get('subnet_mgmt')
        if not subnet_mgmt_id:
            self.fail("subnet_mgmt subnet is not available")

        ext_managed_vl = get_ext_managed_virtual_link("net1", "VL3",
            net1_id)

        network_uuid = self._create_network(neutron_client,
            "external_network")
        subnet_uuid = self._create_subnet(neutron_client, network_uuid)

        # Instantiate vnf
        ext_vl = get_external_virtual_links(
            net0_id, net_mgmt_id, None,
            fixed_addresses=['192.168.120.100'],
            subnet_id=subnet_mgmt_id)

        request_body = self._instantiate_vnf_request("simple",
            vim_id=self.vim_id, ext_vl=ext_vl, ext_managed_vl=ext_managed_vl)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)

        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        vdu_count = len(vnf_instance['instantiatedVnfInfo']
            ['vnfcResourceInfo'])
        self.assertEqual(1, vdu_count)

        # Change external vnf connectivity
        changed_ext_vl = get_external_virtual_links(
            net0_id, network_uuid, None,
            fixed_addresses=['22.22.0.100'],
            subnet_id=subnet_uuid)
        change_ext_conn_req_body = self._change_ext_conn_vnf_request(
            vim_id=self.vim_id, ext_vl=changed_ext_vl)
        before_fixed_ips = self._get_fixed_ips(vnf_instance, request_body)
        self._change_ext_conn_vnf_instance(
            vnf_instance, change_ext_conn_req_body)
        after_fixed_ips = self._get_fixed_ips(vnf_instance, request_body)
        self.assertNotEqual(before_fixed_ips, after_fixed_ips)

        # Get op-occs
        resp, op_occs_info = self._list_op_occs()
        self._assert_occ_list(resp, op_occs_info)

        # Wait for operation state completed
        time.sleep(10)

        # Terminate vnf gracefully with graceful timeout set to 60
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.GRACEFUL,
            'gracefulTerminationTimeout': 60
        }

        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)

        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

EXTENDED_ATTRIBUTES_2_0 = {
    'networks': {
        'v2attrs:something': {'allow_post': False,
                              'allow_put': False,
                              'is_visible': True},
        'v2attrs:something_else': {'allow_post': True,
                                   'allow_put': False,
                                   'is_visible': False},
    }
}


class V2attributes(object):
    def get_name(self):
        return "V2 Extended Attributes Example"

    def get_alias(self):
        return "v2attrs"

    def get_description(self):
        return "Demonstrates extended attributes on V2 core resources"

    def get_namespace(self):
        return "http://docs.openstack.org/ext/examples/v2attributes/api/v1.0"

    def get_updated(self):
        return "2012-07-18T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from cryptography import fernet
from keystoneauth1 import exceptions
from keystoneauth1 import identity
from keystoneauth1 import session
from keystoneclient import client
from oslo_config import cfg
from oslo_log import log as logging


DEFAULT_IDENTITY_VERSION = "v3"
LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class Keystone(object):
    """Keystone module for OpenStack VIM

    Handles identity operations for a given OpenStack
    instance such as version, session and client
    """

    def get_version(self, base_url=None, verify=True):
        try:
            keystone_client = client.Client(auth_url=base_url,
                                            verify=verify)
        except exceptions.ConnectionError:
            raise
        return keystone_client.version

    def get_session(self, auth_plugin, verify):
        ses = session.Session(auth=auth_plugin, verify=verify)
        return ses

    def get_endpoint(self, ses, service_type, region_name=None):
        return ses.get_endpoint(service_type, region_name)

    def initialize_client(self, **kwargs):
        verify = 'True' == kwargs.pop('cert_verify', 'True')
        if 'token' in kwargs:
            auth_plugin = identity.v3.Token(**kwargs)
        else:
            auth_plugin = identity.v3.Password(**kwargs)
        ses = self.get_session(auth_plugin=auth_plugin, verify=verify)
        # note: Using `interface` may be an appropriate way to control
        # the keystone endpoint, e.g., client.Client(DEFAULT_IDENTITY_VERSION,
        # session=ses, interface=interface), but it requires the modification
        # in the DB schema. Thus, use `endpoint_override` for now.
        cli = client.Client(DEFAULT_IDENTITY_VERSION, session=ses,
                            endpoint_override=auth_plugin.auth_url)
        return cli

    @staticmethod
    def create_key_dir(path):
        if not os.access(path, os.F_OK):
            LOG.info('[fernet_tokens] key_repository does not appear to '
                     'exist; attempting to create it')
            try:
                os.makedirs(path, 0o700)
            except OSError:
                LOG.error(
                    'Failed to create [fernet_tokens] key_repository: either '
                    'it already exists or you don\'t have sufficient '
                    'permissions to create it')

    def create_fernet_key(self):
        fernet_key = fernet.Fernet.generate_key()
        fernet_obj = fernet.Fernet(fernet_key)
        return fernet_key, fernet_obj



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# service type constants:
CORE = "CORE"
DUMMY = "DUMMY"
VNFM = "VNFM"
NFVO = "NFVO"
COMMONSERVICES = "COMMONSERVICES"

COMMON_PREFIXES = {
    CORE: "",
    DUMMY: "/dummy_svc",
    VNFM: "",
    NFVO: "",
    COMMONSERVICES: ""
}

# Service operation status constants
ACTIVE = "ACTIVE"
ACK = "ACK"
INACTIVE = "INACTIVE"
PENDING_INSTANTIATE = "PENDING_INSTANTIATE"
PENDING_CREATE = "PENDING_CREATE"
PENDING_UPDATE = "PENDING_UPDATE"
PENDING_DELETE = "PENDING_DELETE"
PENDING_SCALE_IN = "PENDING_SCALE_IN"
PENDING_SCALE_OUT = "PENDING_SCALE_OUT"
PENDING_HEAL = "PENDING_HEAL"
PENDING_TERMINATE = "PENDING_TERMINATE"
PENDING_CHANGE_EXT_CONN = "PENDING_CHANGE_EXT_CONN"
DEAD = "DEAD"
ERROR = "ERROR"
NACK = "NACK"

PENDING_STATUSES = (
    PENDING_INSTANTIATE,
    PENDING_CREATE,
    PENDING_UPDATE,
    PENDING_DELETE,
    PENDING_SCALE_IN,
    PENDING_SCALE_OUT,
    PENDING_HEAL,
    PENDING_TERMINATE,
    PENDING_CHANGE_EXT_CONN,
)
ALL_STATUSES = (
    ACTIVE,
    INACTIVE,
    ERROR,
    *PENDING_STATUSES,
)

POLICY_SCALING = 'tosca.policies.tacker.Scaling'
POLICY_SCALING_ACTIONS = (ACTION_SCALE_OUT,
                          ACTION_SCALE_IN) = ('out', 'in')
POLICY_ACTIONS = {POLICY_SCALING: POLICY_SCALING_ACTIONS}
POLICY_ALARMING = 'tosca.policies.tacker.Alarming'
POLICY_EVENT_ALARMING = 'tosca.policies.tacker.EventAlarming'
VALID_POLICY_TYPES = [POLICY_SCALING, POLICY_ALARMING]
POLICY_RESERVATION = 'tosca.policies.tacker.Reservation'
RESERVATION_POLICY_ACTIONS = ['start_actions',
                              'before_end_actions', 'end_actions']
DEFAULT_ALARM_ACTIONS = ['respawn', 'log', 'log_and_kill', 'notify']

RES_TYPE_VNFD = "vnfd"
RES_TYPE_NSD = "nsd"
RES_TYPE_NS = "ns"
RES_TYPE_VNF = "vnf"
RES_TYPE_VIM = "vim"

RES_EVT_CREATE = "CREATE"
RES_EVT_INSTANTIATE = "INSTANTIATE"
RES_EVT_TERMINATE = "TERMINATE"
RES_EVT_DELETE = "DELETE"
RES_EVT_UPDATE = "UPDATE"
RES_EVT_MONITOR = "MONITOR"
RES_EVT_SCALE = "SCALE"
RES_EVT_NA_STATE = "Not Applicable"
RES_EVT_ONBOARDED = "OnBoarded"
RES_EVT_HEAL = "HEAL"
RES_EVT_MAINTENANCE = [
    "MAINTENANCE", "SCALE_IN", "MAINTENANCE_COMPLETE",
    "PREPARE_MAINTENANCE", "PLANNED_MAINTENANCE", "INSTANCE_ACTION_DONE"
]


VNF_STATUS_TO_EVT_TYPES = {PENDING_CREATE: RES_EVT_CREATE,
                           PENDING_UPDATE: RES_EVT_UPDATE,
                           PENDING_DELETE: RES_EVT_DELETE,
                           PENDING_HEAL: RES_EVT_HEAL}


RES_EVT_CREATED_FLD = "created_at"
RES_EVT_DELETED_FLD = "deleted_at"
RES_EVT_UPDATED_FLD = "updated_at"

TYPE_COMPUTE = "COMPUTE"
TYPE_LINKPORT = "LINKPORT"
TYPE_STORAGE = "STORAGE"
TYPE_VL = "VL"



# Copyright (c) 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import tempfile
import testtools
from unittest import mock
import uuid
import zipfile

from tacker.common import csar_utils
from tacker.common import exceptions
from tacker import context
from tacker.tests import constants
from tacker.tests import utils


class TestCSARUtils(testtools.TestCase):

    def setUp(self):
        super(TestCSARUtils, self).setUp()
        self.context = context.get_admin_context()

    def _get_csar_file_path(self, file_name):
        return os.path.join("./tacker/tests/etc/samples", file_name)

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data(self, mock_extract_csar_zip_file):
        file_path, _ = utils.create_csar_with_unique_vnfd_id(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnfpkg_tosca_vnfd')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_single_yaml(
            self, mock_extract_csar_zip_file):
        file_path, _ = utils.create_csar_with_unique_vnfd_id(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnfpkg_no_meta_single_vnfd')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])

    def _get_csar_zip_from_dir(self, dir_name):
        csar_dir_path = os.path.join('test_csar_utils_data', dir_name)
        unique_name = str(uuid.uuid4())
        csar_temp_dir = os.path.join('/tmp', unique_name)
        self.addCleanup(shutil.rmtree, csar_temp_dir)
        utils.copy_csar_files(csar_temp_dir, csar_dir_path)
        # Copy contents from 'test_csar_utils_common' to 'csar_temp_dir'.
        common_dir_path = ('./tacker/tests/etc/samples/etsi/nfv/'
                           'test_csar_utils_data/test_csar_utils_common')
        common_yaml_file = os.path.join(common_dir_path,
                                        'Definitions/helloworld3_types.yaml')
        shutil.copy(common_yaml_file,
                    os.path.join(csar_temp_dir, 'Definitions/'))

        shutil.copytree(os.path.join(common_dir_path, "TOSCA-Metadata/"),
                        os.path.join(csar_temp_dir, "TOSCA-Metadata/"))
        # Create temporary zip file from 'csar_temp_dir'
        tempfd, tempname = tempfile.mkstemp(suffix=".zip", dir=csar_temp_dir)
        os.close(tempfd)
        zcsar = zipfile.ZipFile(tempname, 'w')

        for (dpath, _, fnames) in os.walk(csar_temp_dir):
            for fname in fnames:
                src_file = os.path.join(dpath, fname)
                dst_file = os.path.relpath(os.path.join(dpath, fname),
                                           csar_temp_dir)
                zcsar.write(src_file, dst_file)
        zcsar.close()
        return tempname

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_in_meta_and_manifest_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        self.assertIsNotNone(vnf_artifacts)
        self.assertIsNotNone(vnf_artifacts[0]['Source'])
        self.assertIsNotNone(vnf_artifacts[0]['Hash'])
        for item in vnf_artifacts:
            flag = item.get('Source').lower().endswith('.img')
            self.assertEqual(flag, False)

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_single_manifest_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_manifest')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        self.assertIsNotNone(vnf_artifacts)
        self.assertIsNotNone(vnf_artifacts[0]['Source'])
        self.assertIsNotNone(vnf_artifacts[0]['Hash'])

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_single_meta_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_meta')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        self.assertIsNotNone(vnf_artifacts)
        self.assertIsNotNone(vnf_artifacts[0]['Source'])
        self.assertIsNotNone(vnf_artifacts[0]['Hash'])

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_meta_in_manifest_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_meta_in_manifest')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        self.assertIsNotNone(vnf_artifacts)
        self.assertIsNotNone(vnf_artifacts[0]['Source'])
        self.assertIsNotNone(vnf_artifacts[0]['Hash'])

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_false_mf_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_false')
        self.addCleanup(os.remove, file_path)
        manifest_path = 'manifest.mf1'
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        msg = (('The file "%(manifest)s" in the CSAR "%(csar)s" does not '
                'contain valid manifest.') %
               {'manifest': manifest_path, 'csar': file_path})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_false_mf_name_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_single_manifest_false_name')
        self.addCleanup(os.remove, file_path)
        manifest_path = 'VNF1.mf'
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        msg = (('The filename "%(manifest)s" is an invalid name.'
                'The name must be the same as the main template '
                'file name.') %
               {'manifest': manifest_path})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_false_hash_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_false_hash')
        self.addCleanup(os.remove, file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        hash_code = '27bbdb25d8f4ed6d07d6f6581b86515e8b2f' \
                    '0059b236ef7b6f50d6674b34f02'
        artifact_path = 'Scripts/install.sh'
        msg = (('The hash "%(hash)s" of artifact file '
                '"%(artifact)s" is an invalid value.') %
               {'hash': hash_code, 'artifact': artifact_path})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_missing_key_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_missing_key')
        self.addCleanup(os.remove, file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        key_name = sorted(['Algorithm'])
        msg = (('One of the artifact information may not have '
                'the key("%(key)s")') % {'key': key_name})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_missing_value_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_missing_value')
        self.addCleanup(os.remove, file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        key_name = 'Algorithm'
        msg = (('One of the artifact information may not have '
                'the key value("%(key)s")') % {'key': key_name})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_false_source_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_false_source')
        self.addCleanup(os.remove, file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        artifact_path = 'Scripts/install.s'
        msg = (('The path("%(artifact_path)s") of '
                'artifact Source is an invalid value.') %
               {'artifact_path': artifact_path})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_false_algorithm_with_vnf_artifact(
            self, mock_extract_csar_zip_file):
        file_path = utils.create_csar_with_unique_artifact(
            './tacker/tests/etc/samples/etsi/nfv/'
            'sample_vnf_package_csar_in_meta_and_manifest_false_algorithm')
        self.addCleanup(os.remove, file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        algorithm = 'sha-255'
        artifact_path = 'Scripts/install.sh'
        msg = (('The algorithm("%(algorithm)s") of '
                'artifact("%(artifact_path)s") is '
                'an invalid value.') %
               {'algorithm': algorithm,
                'artifact_path': artifact_path})
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_without_instantiation_level(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_without_instantiation_level')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                          csar_utils.load_csar_data,
                          self.context, constants.UUID, file_path)
        msg = ('Policy of type'
               ' "tosca.policies.nfv.InstantiationLevels is not defined.')
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_invalid_instantiation_level(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_invalid_instantiation_level')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        levels = ['instantiation_level_1', 'instantiation_level_2']
        msg = ("Level(s) instantiation_level_3 not found in "
               "defined levels %s") % ",".join(sorted(levels))
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_invalid_default_instantiation_level(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_with_invalid_default_instantiation_level')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        levels = ['instantiation_level_1', 'instantiation_level_2']
        msg = ("Level instantiation_level_3 not found in "
               "defined levels %s") % ",".join(sorted(levels))
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_without_vnfd_info(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_without_vnfd_info')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        self.assertEqual("VNF properties are mandatory", exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_artifacts_and_without_sw_image_data(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_without_sw_image_data')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        msg = ('Node property "sw_image_data" is missing for '
               'artifact sw_image for node VDU1.')
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_multiple_sw_image_data(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_with_multiple_sw_image_data')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        msg = ('artifacts of type "tosca.artifacts.nfv.SwImage"'
               ' is added more than one time for node VDU1.')
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_csar_with_missing_sw_image_data_in_main_template(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_with_missing_sw_image_data_in_main_template')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        msg = ('Node property "sw_image_data" is missing for'
               ' artifact sw_image for node VDU1.')
        self.assertEqual(msg, exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_without_flavour_info(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir('csar_without_flavour_info')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        self.assertEqual("No VNF flavours are available", exc.format_message())

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_without_flavour_info_in_main_template(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_without_flavour_info_in_main_template')
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, file_path)
        self.assertEqual("No VNF flavours are available",
                         exc.format_message())

    @mock.patch.object(os, 'remove')
    @mock.patch.object(shutil, 'rmtree')
    def test_delete_csar_data(self, mock_rmtree, mock_remove):
        csar_utils.delete_csar_data(constants.UUID)
        mock_rmtree.assert_called()
        mock_remove.assert_called()

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_without_policies(
            self, mock_extract_csar_zip_file):
        file_path = self._get_csar_zip_from_dir(
            'csar_without_policies')
        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertIsNone(flavours[0].get('instantiation_levels'))
        self.assertEqual(vnf_data['descriptor_version'], '1.0')

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_with_artifacts_short_notation_without_sw_image_data(
            self, mock_extract_csar_zip_file):
        file_path = "./tacker/tests/etc/samples/etsi/nfv/" \
                    "csar_short_notation_for_artifacts_without_sw_image_data"
        zip_name, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, zip_name)
        msg = ('Node property "sw_image_data" is missing for'
               ' artifact sw_image for node VDU1.')
        self.assertEqual(msg, exc.format_message())
        os.remove(zip_name)

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_artifacts_short_notation(
            self, mock_extract_csar_zip_file):
        file_path = "./tacker/tests/etc/samples/etsi/nfv/" \
                    "csar_with_short_notation_for_artifacts"
        zip_name, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)

        vnf_data, flavours, vnf_artifacts = csar_utils.load_csar_data(
            self.context, constants.UUID, zip_name)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        os.remove(zip_name)

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_multiple_sw_image_data_with_short_notation(
            self, mock_extract_csar_zip_file):

        file_path = "./tacker/tests/etc/samples/etsi/nfv/" \
                    "csar_multiple_sw_image_data_with_short_notation"
        zip_name, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)
        exc = self.assertRaises(exceptions.InvalidCSAR,
                                csar_utils.load_csar_data,
                                self.context, constants.UUID, zip_name)
        msg = ('artifacts of type "tosca.artifacts.nfv.SwImage"'
               ' is added more than one time for node VDU1.')
        self.assertEqual(msg, exc.format_message())
        os.remove(zip_name)

    @mock.patch('tacker.common.csar_utils.extract_csar_zip_file')
    def test_load_csar_data_with_unit_conversion(
            self, mock_extract_csar_zip_file):
        file_path, _ = utils.create_csar_with_unique_vnfd_id(
            './tacker/tests/etc/samples/etsi/nfv/sample_vnfpkg_tosca_vnfd')
        self.addCleanup(os.remove, file_path)
        vnf_data, flavours, vnf_artifact = csar_utils.load_csar_data(
            self.context, constants.UUID, file_path)
        self.assertEqual(vnf_data['descriptor_version'], '1.0')
        self.assertEqual(vnf_data['vnfm_info'], ['Tacker'])
        self.assertEqual(flavours[0]['flavour_id'], 'simple')
        self.assertIsNotNone(flavours[0]['sw_images'])
        # 'size', 'min_disk' and 'min_ram' values from sample VNFD will
        # be converted to Bytes
        self.assertEqual(flavours[0]['sw_images'][0]['min_disk'], 1000000000)
        self.assertEqual(flavours[0]['sw_images'][0]['size'], 1879048192)
        self.assertEqual(flavours[0]['sw_images'][1]['min_disk'], 2000000000)
        self.assertEqual(flavours[0]['sw_images'][1]['size'], 2000000000)
        self.assertEqual(flavours[0]['sw_images'][1]['min_ram'], 8590458880)

    def test_extract_csar_zip_file_for_file_permissions(self):
        dir_location = tempfile.mkdtemp()
        file_obj = tempfile.NamedTemporaryFile(dir=dir_location)
        file_path = file_obj.name
        file_name = file_path.split("/")[3]
        os.chmod(file_path, 0o755)
        self.addCleanup(shutil.rmtree, dir_location)
        initial_cwd = os.getcwd()
        os.chdir(dir_location)
        zip_file_path = dir_location + '/' + "test_extract_csar_zip_file.zip"
        with zipfile.ZipFile(zip_file_path, 'w') as zip:
            zip.write(file_name)
        self.assertEqual(True, os.path.exists(zip_file_path))
        zip_extract_path = dir_location + "/zip_extract"
        extract_file_path = zip_extract_path + "/" + file_name
        os.chdir(initial_cwd)
        csar_utils.extract_csar_zip_file(zip_file_path, zip_extract_path)
        status = os.stat(extract_file_path)
        permission = oct(status.st_mode)[-3:]
        self.assertEqual('755', permission)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from tacker.policies import base


VNFLCM = 'os_nfv_orchestration_api:vnf_instances:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'api_versions',
        check_str=base.RULE_ANY,
        description="Get API Versions.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnflcm/v1/api_versions'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'create',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Creates vnf instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_instances'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'instantiate',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Instantiate vnf instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}/instantiate'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'show',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Query an Individual VNF instance.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'terminate',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Terminate a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}/terminate'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'heal',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Heal a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}/heal'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'scale',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Scale a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}/scale'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'show_lcm_op_occs',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Query an Individual VNF LCM operation occurrence",
        operations=[
            {
                'method': 'GET',
                'path': '/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'list_lcm_op_occs',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Query VNF LCM operation occurrence",
        operations=[
            {
                'method': 'GET',
                'path': '/vnflcm/v1/vnf_lcm_op_occs'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'index',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Query VNF instances.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnflcm/v1/vnf_instances'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'delete',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Delete an Individual VNF instance.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'update_vnf',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Update an Individual VNF instance.",
        operations=[
            {
                'method': 'PATCH',
                'path': '/vnflcm/v1/vnf_instances/{vnfInstanceId}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'rollback',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Rollback a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/rollback'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'cancel',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Cancel a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/cancel'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'fail',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Fail a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/fail'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'retry',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Retry a VNF instance.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/retry'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=VNFLCM % 'change_ext_conn',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Change external VNF connectivity.",
        operations=[
            {
                'method': 'POST',
                'path':
                    '/vnflcm/v1/vnf_instances/{vnfInstanceId}/change_ext_conn'
            }
        ]
    ),
]


def list_rules():
    return rules



# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""rename the field a to b

Revision ID: 85c06a0714b7
Revises: d6ae359ab0d6
Create Date: 2022-01-14 06:25:30.427454

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '85c06a0714b7'
down_revision = 'd6ae359ab0d6'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):

    op.alter_column('vnf_lcm_subscriptions',
                    'subscription_authentication',
                    new_column_name='authentication',
                    existing_type=sa.JSON(), nullable=True)




# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import ddt

from tacker import context
from tacker.db import api as sqlalchemy_api
from tacker.db.nfvo import nfvo_db
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel

get_engine = sqlalchemy_api.get_engine


@ddt.ddt
class TestVnfd(SqlTestCase):

    def setUp(self):
        super(TestVnfd, self).setUp()
        self.context = context.get_admin_context()
        self.vims = nfvo_db.Vim(**fakes.vim_data)
        self.engine = get_engine()
        self.conn = self.engine.connect()
        self.vim = nfvo_db.Vim()
        self.vnfd_data = {
            "id": uuidsentinel.id,
            "tenant_id": uuidsentinel.vnfd_id,
            "name": "name",
            "description": "description",
            "mgmt_driver": "mgmt_driver",
            "deleted_at": datetime.datetime.now()
        }
        self.vnfd_object = objects.vnfd.Vnfd(
            context=self.context, **self.vnfd_data)

    def test_obj_from_db_obj(self):
        db_obj = {"key": "value"}
        result = self.vnfd_object.obj_from_db_obj(
            self.context, db_obj)
        self.assertIsInstance(result, objects.vnfd.Vnfd)

    def test_destroy(self):
        self.vnfd_object.destroy(self.vnfd_object.id)

    def test_delete(self):
        self.vnfd_object.delete(self.vnfd_object.id)

    def test_check_vnfd_attribute(self):
        result = self.vnfd_object.check_vnfd(
            self.vnfd_object.id)
        self.assertEqual(result, "FALSE")
